package com.bif.timecard.invoice;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import java.net.URLConnection;
import java.text.DecimalFormat;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

import javax.sql.RowSet;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.ClientAnchor;
import org.apache.poi.ss.usermodel.CreationHelper;
import org.apache.poi.ss.usermodel.DataFormat;
import org.apache.poi.ss.usermodel.Drawing;
import org.apache.poi.ss.usermodel.FillPatternType;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.HorizontalAlignment;
import org.apache.poi.ss.usermodel.IndexedColors;
import org.apache.poi.ss.usermodel.Picture;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.VerticalAlignment;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.util.IOUtils;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import com.bif.timecard.loader.DAOTimeCard;

/**
 * <p>Instances of this class generate an invoice that is acceptable to the customer.</p>
 * 
 * @author <a href="mailto:roderick.barnes@biftechnologies.com">Roderick L. Barnes, Sr.</a>
 */
public class InvoiceGenerator {
	
	private int intLoopIndex = 0;
	
	/**
	 * For the generation of an invoice this attribute of the
	 * the instance helps filter time and attendance data.
	 */
	private String stringBillingPeriodID = null;
	private DAOTimeCard daoTimeCard = new DAOTimeCard();
	private Font font = null;
	
	private ArrayList<String> arrayListOfTotalQuantityOfItemsAddress = new ArrayList<String>();
	
	private ArrayList<String> arrayListOfTotalInvoiceAmountAddress = new ArrayList<String>();
	
	/**
	 * Sets the billing period ID for an instance of {@code InvoiceGenerator}.
	 * @param stringBillingPeriodIDNew billing period identifier
	 */
	public void setBillingPeriodID(String stringBillingPeriodIDNew) {
		this.stringBillingPeriodID = stringBillingPeriodIDNew;
	}
	
	/**
	 * Returns the billing period ID for this instance of {@code InvoiceGenerator}.
	 * @return the billing period ID for this instance of {@code InvoiceGenerator}.
	 */
	public String getBillingPeriodID() {
		return this.stringBillingPeriodID;
	}
	
	public void setFont(Font fontNew) {
		this.font = fontNew;
	}
	
	public Font getFont() {
		return this.font;
	}
	
	public ArrayList<String> getQuantityOfItemsAddressList() {
		return this.arrayListOfTotalQuantityOfItemsAddress;
	}
	
	public ArrayList<String> getTotalInvoiceAmountAddressList() {
		return this.arrayListOfTotalInvoiceAmountAddress;
	}
	
	/**
	 * Returns a reference to the data access object for the BrainJack Time Card
	 * database.
	 * @return a reference to the data access object for the BrainJack Time Card
	 * database.
	 */
	private DAOTimeCard getDAOTimeCard() {
		return this.daoTimeCard;
	}
	
	public static double getTruncatedDouble(String stringFormat, double doubleNumber) {
		DecimalFormat decimalFormat = new DecimalFormat(stringFormat);
		
		
		String stringDouble = decimalFormat.format(doubleNumber);
		/**
		 * Remove everything beyond the two digits after the decimal.
		 */
		String stringCopy = stringDouble;
		stringDouble = stringDouble.substring(0,stringDouble.length() - 2);
		return Double.parseDouble(stringDouble);
	}
	
	/**
	 * Create the header for the details section. This will include the titles "Direct Labor CLIN",
	 * "Grand Total", "Employee", etc.
	 * @param sheet the sheet to which the header information will be added.
	 */
	public void createHeaderForDetails(Sheet sheet) {
		CellStyle cellStyle = sheet.getWorkbook().createCellStyle();
		cellStyle.setAlignment(HorizontalAlignment.CENTER);
		Font font = sheet.getWorkbook().createFont();
		font.setFontHeightInPoints((short) 18);
		font.setBold(true);
		font.setFontName(this.getFont().getFontName());
		font.setFontHeightInPoints(this.getFont().getFontHeightInPoints());
		cellStyle.setFont(this.getFont());
		
		int intRowOfHeaderForDetailsStart = sheet.getPhysicalNumberOfRows() + 1;
		
		/*
		 * Create header row for invoice details.
		 */
		Row rowDetailsSuperHeader = sheet.createRow(intRowOfHeaderForDetailsStart++);
		
		Cell cellDirectLabor = rowDetailsSuperHeader.createCell(0);
		cellDirectLabor.setCellStyle(cellStyle);
		cellDirectLabor.setCellValue("Direct Labor CLIN 0001");
		
		Cell cellGrandTotal = rowDetailsSuperHeader.createCell(3);
		cellGrandTotal.setCellStyle(cellStyle);
		cellGrandTotal.setCellValue("Grand Total");
		sheet.addMergedRegion(new CellRangeAddress(intRowOfHeaderForDetailsStart - 1, intRowOfHeaderForDetailsStart - 1, 3, 5));
		
		CellStyle detailsHeaderCellStyle = sheet.getWorkbook().createCellStyle();
		detailsHeaderCellStyle.setAlignment(HorizontalAlignment.CENTER);
		detailsHeaderCellStyle.setFont(this.getFont());

		Row rowDetailsHeader = sheet.createRow(intRowOfHeaderForDetailsStart++);
		Cell cellCLINDescriptionHeader = rowDetailsHeader.createCell(0);
		cellCLINDescriptionHeader.setCellStyle(detailsHeaderCellStyle);
		cellCLINDescriptionHeader.setCellValue("CLIN/Description");
		
		Cell cellEmployeeOrdinalHeader = rowDetailsHeader.createCell(1);
		cellEmployeeOrdinalHeader.setCellStyle(detailsHeaderCellStyle);
		cellEmployeeOrdinalHeader.setCellValue("#");
		
		Cell cellEmployeeHeader = rowDetailsHeader.createCell(2);
		cellEmployeeHeader.setCellStyle(detailsHeaderCellStyle);
		cellEmployeeHeader.setCellValue("Employee");
		
		Cell cellEmployeeLaborCategoryCount = rowDetailsHeader.createCell(3);
		cellEmployeeLaborCategoryCount.setCellStyle(detailsHeaderCellStyle);
		cellEmployeeLaborCategoryCount.setCellValue("Employees Per Category");
		
		Cell cellQuantity = rowDetailsHeader.createCell(4);
		cellQuantity.setCellStyle(detailsHeaderCellStyle);
		cellQuantity.setCellValue("Quantity");
		
		Cell cellUnitPrice = rowDetailsHeader.createCell(5);
		cellUnitPrice.setCellStyle(detailsHeaderCellStyle);
		cellUnitPrice.setCellValue("Unit Price");
		
		Cell cellAmount = rowDetailsHeader.createCell(6);
		cellAmount.setCellStyle(detailsHeaderCellStyle);
		cellAmount.setCellValue("Amount");		
	}
	
	private List<InvoiceLaborCategoryGroup> getLaborCategoryGroups(RowSet rowSet) throws Exception {
		List<InvoiceLaborCategoryGroup> listOfInvoiceLaborCategoryGroup = new ArrayList<InvoiceLaborCategoryGroup>();
		InvoiceLaborCategoryGroup invoiceLaborCategoryGroup = null;
		
		String stringCurrentLaborCategory = "";
		rowSet.beforeFirst();
		while(rowSet.next()) {
			/**
			 * If the current labor category does not match the last one we have a new grouping.
			 */
			if (!rowSet.getString("labor_category").equalsIgnoreCase(stringCurrentLaborCategory)) {
				stringCurrentLaborCategory = rowSet.getString("labor_category");
				
				String stringLaborCategoryID = stringCurrentLaborCategory.split("-")[0].trim();
				String stringLaborCategoryName = stringCurrentLaborCategory.split("-")[1].trim();
				invoiceLaborCategoryGroup = new InvoiceLaborCategoryGroup();
				invoiceLaborCategoryGroup.setLaborCategoryID(stringLaborCategoryID);
				invoiceLaborCategoryGroup.setLaborCategoryName(stringLaborCategoryName);
				
				listOfInvoiceLaborCategoryGroup.add(invoiceLaborCategoryGroup);
			}
			
			EmployeeHoursSummary employeeHoursSummary = new EmployeeHoursSummary();
			String stringEmployeeID = rowSet.getString("employee").split("\\(")[1].trim();
			stringEmployeeID = stringEmployeeID.replace(")", "");
			String stringNameTemp = rowSet.getString("employee").split("\\(")[0].trim();
			String stringLastName = stringNameTemp.split(",")[0];
			String stringFirstName = stringNameTemp.split(",")[1];
			employeeHoursSummary.setEmployeeID(stringEmployeeID);
			employeeHoursSummary.setFirstName(stringFirstName);
			employeeHoursSummary.setLastName(stringLastName);
			employeeHoursSummary.setNonProfitHours(rowSet.getDouble("non_profit_quantity"));
			employeeHoursSummary.setProfitHours(rowSet.getDouble("quantity"));
			employeeHoursSummary.setNonProfitHourlyRate(rowSet.getDouble("non_profit_unit_cost"));
			employeeHoursSummary.setProfitHourlyRate(rowSet.getDouble("unit_cost"));
			
			invoiceLaborCategoryGroup.getEmployeeHoursSummary().add(employeeHoursSummary);
		}
		
		return listOfInvoiceLaborCategoryGroup;
	}
	
	public void createSectionHeader(Workbook workbook, Sheet sheet, DTOInvoiceMetaData dtoInvoiceMetaData) throws Exception {
		
		CellStyle cellStyle = sheet.getWorkbook().createCellStyle();
		cellStyle.setWrapText(true);
		cellStyle.setVerticalAlignment(VerticalAlignment.TOP);
		cellStyle.setFont(this.getFont());
		
		CellStyle cellStyleForBoldedFields = sheet.getWorkbook().createCellStyle();
		Font font = sheet.getWorkbook().createFont();
		font.setFontHeightInPoints((short) 18);
		font.setBold(true);
		font.setFontName(this.getFont().getFontName());
		
		cellStyleForBoldedFields.setFont(font);
		cellStyleForBoldedFields.setAlignment(HorizontalAlignment.LEFT);

		DataFormat dataFormat = sheet.getWorkbook().createDataFormat();
		CellStyle cellStyleForInvoiceSectionTotalHours = sheet.getWorkbook().createCellStyle();
		cellStyleForInvoiceSectionTotalHours.setFont(font);
		cellStyleForInvoiceSectionTotalHours.setAlignment(HorizontalAlignment.RIGHT);
		cellStyleForInvoiceSectionTotalHours.setDataFormat(dataFormat.getFormat("###,##0.00"));
		CellStyle cellStyleForInvoiceSectionTotalInvoiceAmount = sheet.getWorkbook().createCellStyle();
		cellStyleForInvoiceSectionTotalInvoiceAmount.setFont(font);
		cellStyleForInvoiceSectionTotalInvoiceAmount.setAlignment(HorizontalAlignment.RIGHT);
		cellStyleForInvoiceSectionTotalInvoiceAmount.setDataFormat(dataFormat.getFormat("$##,###,##0.00"));
		
		/**
		 * Setup cell style for period, payment terms, and invoice date
		 */
		CellStyle cellStyleForPeriod = sheet.getWorkbook().createCellStyle();
		cellStyleForPeriod.setFont(this.getFont());
		cellStyleForPeriod.setAlignment(HorizontalAlignment.RIGHT);
		CellStyle cellStyleForPaymentTerms = sheet.getWorkbook().createCellStyle();
		cellStyleForPaymentTerms.setFont(this.getFont());
		cellStyleForPaymentTerms.setAlignment(HorizontalAlignment.RIGHT);
		CellStyle cellStyleForDateOfInvoice = sheet.getWorkbook().createCellStyle();
		cellStyleForDateOfInvoice.setFont(this.getFont());
		cellStyleForDateOfInvoice.setAlignment(HorizontalAlignment.RIGHT);
		CellStyle cellStyleForInvoiceNumber = sheet.getWorkbook().createCellStyle();
		cellStyleForInvoiceNumber.setFont(font);
		cellStyleForInvoiceNumber.setAlignment(HorizontalAlignment.RIGHT);
		
		/*
		 * Step 1: Put XSI logo in upper left corner.
		 */
		URL url = new URL(dtoInvoiceMetaData.getLogoLocation());
		URLConnection urlConnection = url.openConnection();
		
		CreationHelper creationHelper = workbook.getCreationHelper();
		
	    InputStream inputStream = urlConnection.getInputStream();
	    byte[] bytes = IOUtils.toByteArray(inputStream);
	    int pictureIdx = workbook.addPicture(bytes, Workbook.PICTURE_TYPE_JPEG);
	    inputStream.close();
	    
	    /*
	     * Step 1.1: Create the drawing patriarch.  This is the top level container for all shapes. 
	     */
	    Drawing<?> drawing = sheet.createDrawingPatriarch();
	    /*
	     * Step 1.2: Add a picture shape
	     */
	    ClientAnchor anchor = creationHelper.createClientAnchor();
	    /*
	     * Step 1.3: Set top-left corner of the picture. Subsequent calls of Picture#resize() will
	     * operate relative to this coordinate.
	     */
	    anchor.setCol1(0);
	    anchor.setRow1(0);
	    Picture pict = drawing.createPicture(anchor, pictureIdx);
	    pict.resize(1.0, 7.0);
		
		/*
		 * Step 2: Place DUNS, Requisition Number, Contract Number, Order Number, Task Order Number, and FARS
		 */
	    CellStyle cellStyleForHeaderFields = sheet.getWorkbook().createCellStyle();
	    cellStyleForHeaderFields.setFont(this.getFont());
	    
		Row rowHeader01 = sheet.createRow(0);
		Cell cellDUNS = rowHeader01.createCell(2);
		cellDUNS.setCellStyle(cellStyleForHeaderFields);
		cellDUNS.setCellValue("DUNS Number: " + dtoInvoiceMetaData.getDUNSNumber());
		Cell cellFrom = rowHeader01.createCell(5);
		cellFrom.setCellStyle(cellStyleForHeaderFields);
		cellFrom.setCellValue("Period:");
		Cell cellFromValue = rowHeader01.createCell(6);
		cellFromValue.setCellStyle(cellStyleForPeriod);
		cellFromValue.setCellValue(dtoInvoiceMetaData.getBillingPeriod());
		
		Row rowHeader02 = sheet.createRow(1);
		Cell cellRequisitionNumber = rowHeader02.createCell(2);
		cellRequisitionNumber.setCellStyle(cellStyleForHeaderFields);
		cellRequisitionNumber.setCellValue("Requisition Number: " + dtoInvoiceMetaData.getRequisitionNumber());
		
		Row rowHeader03 = sheet.createRow(2);
		Cell cellContractNumber = rowHeader03.createCell(2);
		cellContractNumber.setCellStyle(cellStyleForHeaderFields);
		cellContractNumber.setCellValue("Contract Number: " + dtoInvoiceMetaData.getContractNumber());
		Cell cellInvoiceNumber = rowHeader03.createCell(5);
		cellInvoiceNumber.setCellStyle(cellStyleForBoldedFields);
		cellInvoiceNumber.setCellValue("Invoice#");
		Cell cellInvoiceNumberValue = rowHeader03.createCell(6);
		cellInvoiceNumberValue.setCellStyle(cellStyleForInvoiceNumber);
		cellInvoiceNumberValue.setCellValue(dtoInvoiceMetaData.getInvoiceNumber());

		Row rowHeader04 = sheet.createRow(3);
		Cell cellOrderNumber = rowHeader04.createCell(2);
		cellOrderNumber.setCellStyle(cellStyleForBoldedFields);
		cellOrderNumber.setCellValue("Order Number: " + dtoInvoiceMetaData.getOrderNumber());
		Cell cellPaymentTerms = rowHeader04.createCell(5);
		cellPaymentTerms.setCellStyle(cellStyleForHeaderFields);
		cellPaymentTerms.setCellValue("Payment Terms:");
		Cell cellPaymentTermsValue = rowHeader04.createCell(6);
		cellPaymentTermsValue.setCellStyle(cellStyleForPaymentTerms);
		cellPaymentTermsValue.setCellValue(dtoInvoiceMetaData.getPaymentTerms());

		Row rowHeader05 = sheet.createRow(4);
		Cell cellTaskOrderNumber = rowHeader05.createCell(2);
		cellTaskOrderNumber.setCellStyle(cellStyleForHeaderFields);
		cellTaskOrderNumber.setCellValue("Task Order Number: " + dtoInvoiceMetaData.getTaskOrderNumber());
		
		/*
		 * Step 3: Place From, To, Invoice#, Payment Terms, Date of Invoice, Total Quantity (Hours), Total Amount (Dollars)
		 */
		Cell cellDateOfInvoice = rowHeader05.createCell(5);
		cellDateOfInvoice.setCellStyle(cellStyleForHeaderFields);
		cellDateOfInvoice.setCellValue("Date of Invoice:");
		Cell cellDateOfInvoiceValue = rowHeader05.createCell(6);
		cellDateOfInvoiceValue.setCellStyle(cellStyleForDateOfInvoice);
		cellDateOfInvoiceValue.setCellValue(dtoInvoiceMetaData.getDateOfInvoice());
		
		Row rowHeader06 = sheet.createRow(5);
		Cell cellTotalQuantityOfItems = rowHeader06.createCell(5);
		cellTotalQuantityOfItems.setCellStyle(cellStyleForBoldedFields);
		cellTotalQuantityOfItems.setCellValue("Total Quantity of Items:");
		Cell cellTotalQuantityOfItemsValue = rowHeader06.createCell(6);
		cellTotalQuantityOfItemsValue.setCellStyle(cellStyleForInvoiceSectionTotalHours);
		cellTotalQuantityOfItemsValue.setCellValue(dtoInvoiceMetaData.getTotalHours());
		
		Row rowHeader07 = sheet.createRow(6);
		Cell cellTotalInvoiceAmount = rowHeader07.createCell(5);
		cellTotalInvoiceAmount.setCellStyle(cellStyleForBoldedFields);
		cellTotalInvoiceAmount.setCellValue("Total Invoice Amount:");
		Cell cellTotalInvoiceAmountValue = rowHeader07.createCell(6);
		cellTotalInvoiceAmountValue.setCellStyle(cellStyleForInvoiceSectionTotalInvoiceAmount);
		cellTotalInvoiceAmountValue.setCellValue(dtoInvoiceMetaData.getTotalInvoiceAmount());
		
		/*
		 * Step 4: Place Xtreme Address and Bill To.
		 */
		Row rowHeader08 = sheet.createRow(8);
		Cell cellXtremeAddress = rowHeader08.createCell(0);
		cellXtremeAddress.setCellValue(dtoInvoiceMetaData.getVendorAddress());
		cellXtremeAddress.setCellStyle(cellStyle);
		//sheet.addMergedRegion(new CellRangeAddress(8, 12, 0, 0));
		Cell cellBillTo = rowHeader08.createCell(2);
		cellBillTo.setCellValue("Bill To:\n" + dtoInvoiceMetaData.getStringCustomerBillToAddress());
		cellBillTo.setCellStyle(cellStyle);
		//sheet.addMergedRegion(new CellRangeAddress(8, 15, 2, 2));	
	}
	
	public void createSectionFooter(Sheet sheet, DTOInvoiceMetaData dtoInvoiceMetaData) throws Exception {
		/**
		 * Step 1: Setup the style objects required for later in the production
		 * of the footer. 
		 */
		
		CellStyle cellStyle = sheet.getWorkbook().createCellStyle();
		cellStyle.setWrapText(true);
		cellStyle.setVerticalAlignment(VerticalAlignment.TOP);
		cellStyle.setFont(this.getFont());
		
		int intNumberOfReportRows = sheet.getPhysicalNumberOfRows() + 3;
		
		/**
		 * Step 2: Place REMIT TO, Remit Contact Info
		 */
		Row rowFooter01 = sheet.createRow(intNumberOfReportRows++);
		Cell cellRemitToAddress = rowFooter01.createCell(0);
		cellRemitToAddress.setCellStyle(cellStyle);
		cellRemitToAddress.setCellValue("REMIT TO: " + dtoInvoiceMetaData.getRemitToAddress());
		Cell cellRemitContactInformation = rowFooter01.createCell(2);
		cellRemitContactInformation.setCellStyle(cellStyle);
		cellRemitContactInformation.setCellValue("Remittance of Contact Information\n" + dtoInvoiceMetaData.getRemitContactInformation());
		
		/*
		 * Step 3: Remit route
		 */
		Row rowFooter02 = sheet.createRow(intNumberOfReportRows + 2);
		Cell cellRemitMethod = rowFooter02.createCell(0);
		cellRemitMethod.setCellStyle(cellStyle);
		cellRemitMethod.setCellValue(dtoInvoiceMetaData.getRemitMethodInformation());
	}
	
	public byte[] generateInvoice() throws Exception {
		/**
		 * Step 0: Demand that we have a billing period before attempting the generation process.
		 */
		if (this.getBillingPeriodID() == null && this.getBillingPeriodID().trim().length() > 0) {
			throw new Exception("Billing period has not been set. Set this first.");
		}
		
		/**
		 * Step 1: Create a workbook for the invoice.
		 */
		System.out.println("Creating workbook for invoice spreadsheet.");
		Workbook workbook = new XSSFWorkbook();
		
		/**
		 * Step 1.1: Setup the font for the spreadsheets.
		 */
		Font fontTemp = workbook.createFont();
		fontTemp.setFontName("Times New Roman");
		fontTemp.setFontHeightInPoints((short) 18);
		this.setFont(fontTemp);
		
		/**
		 * Step 1.2: Create summary and profit formatting objects.
		 */
		DataFormat dataFormat = workbook.createDataFormat();
		CellStyle cellStyleHeader = workbook.createCellStyle();
		cellStyleHeader.setFont(this.getFont());
		CellStyle cellStyleHours = workbook.createCellStyle();
		cellStyleHours.setDataFormat(dataFormat.getFormat("###,##0.00"));
		cellStyleHours.setFont(this.getFont());
		CellStyle cellStyleRate = workbook.createCellStyle();
		cellStyleRate.setDataFormat(dataFormat.getFormat("##0.00"));
		cellStyleRate.setFont(this.getFont());
		CellStyle cellStyleExtendedCost = workbook.createCellStyle();
		cellStyleExtendedCost.setDataFormat(dataFormat.getFormat("#,###,##0.00"));
		cellStyleExtendedCost.setFont(this.getFont());
		
		/**
		 * Step 1.3: Create summary and profit formatting objects.
		 */
		CellStyle cellStyleNonProfitEmployeeName = workbook.createCellStyle();
		cellStyleNonProfitEmployeeName.setFillForegroundColor(IndexedColors.LIGHT_CORNFLOWER_BLUE.index);
		cellStyleNonProfitEmployeeName.setFillPattern(FillPatternType.SOLID_FOREGROUND);
		cellStyleNonProfitEmployeeName.setFont(this.getFont());

		CellStyle cellStyleNonProfitHours = workbook.createCellStyle();
		cellStyleNonProfitHours.setDataFormat(dataFormat.getFormat("###,##0.00"));
		cellStyleNonProfitHours.setFillForegroundColor(IndexedColors.LIGHT_CORNFLOWER_BLUE.index);
		cellStyleNonProfitHours.setFillPattern(FillPatternType.SOLID_FOREGROUND);
		cellStyleNonProfitHours.setFont(this.getFont());
		
		CellStyle cellStyleNonProfitRate = workbook.createCellStyle();
		cellStyleNonProfitRate.setFillForegroundColor(IndexedColors.LIGHT_CORNFLOWER_BLUE.index);
		cellStyleNonProfitRate.setFillPattern(FillPatternType.SOLID_FOREGROUND);
		cellStyleNonProfitRate.setDataFormat(dataFormat.getFormat("##0.00"));
		cellStyleNonProfitRate.setFont(this.getFont());
		
		CellStyle cellStyleNonProfitExtendedCost = workbook.createCellStyle();
		cellStyleNonProfitExtendedCost.setDataFormat(dataFormat.getFormat("#,###,##0.00"));
		cellStyleNonProfitExtendedCost.setFillForegroundColor(IndexedColors.LIGHT_CORNFLOWER_BLUE.index);
		cellStyleNonProfitExtendedCost.setFillPattern(FillPatternType.SOLID_FOREGROUND);
		cellStyleNonProfitExtendedCost.setFont(this.getFont());
		
		Font fontAggregateValues = workbook.createFont();
		fontAggregateValues.setFontName("Times New Roman");
		fontAggregateValues.setFontHeightInPoints((short) 18);
		fontAggregateValues.setBold(true);
		
		CellStyle cellStyleTotalInvoiceHours = workbook.createCellStyle();
		cellStyleTotalInvoiceHours.setDataFormat(dataFormat.getFormat("#,###,##0.00"));
		cellStyleTotalInvoiceHours.setFont(fontAggregateValues);
		cellStyleTotalInvoiceHours.setAlignment(HorizontalAlignment.RIGHT);
		
		CellStyle cellStyleTotalInvoiceAmount = workbook.createCellStyle();
		cellStyleTotalInvoiceAmount.setDataFormat(dataFormat.getFormat("$#,###,##0.00"));
		cellStyleTotalInvoiceAmount.setFont(fontAggregateValues);
		cellStyleTotalInvoiceAmount.setAlignment(HorizontalAlignment.RIGHT);

		/**
		 * Step 2: Create a rowset for the sections of the invoice.
		 */
		System.out.println("Getting rowset of sections for invoice spreadsheet.");
		RowSet rowSetSections = this.getDAOTimeCard().getSections(this.getBillingPeriodID());
		rowSetSections.beforeFirst();
		while (rowSetSections.next()) {
			System.out.println("\tInvoice Section: " + rowSetSections.getString(2));
		}
		
		/*
		 * Step 3: Create the Invoice metadata object and set the metadata.
		 */
		DTOInvoiceMetaData dtoInvoiceMetaData = this.getDAOTimeCard().getInvoiceMetaData(this.getBillingPeriodID());
		System.out.println("DTOInvoiceMetaData: " + dtoInvoiceMetaData);
		
		/*
		 * Step 3.1: Set the billing period for the invoice.
		 */
		dtoInvoiceMetaData.setBillingPeriod(this.getBillingPeriodID());
		
		/*
		 * Step 3.2: Set the date of the invoice.
		 */
		LocalDate localDateNow = LocalDate.now();
		DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy.MM.dd");
		String stringCurrentDate = localDateNow.format(dateTimeFormatter);
		System.out.println("Setting Invoice Date: " + stringCurrentDate);
		dtoInvoiceMetaData.setDateOfInvoice(stringCurrentDate);
		
		CellStyle cellStyle = workbook.createCellStyle();
		cellStyle.setFont(this.getFont());
		
		/*
		 * Step 4: Iterate through invoice sections.
		 */
		rowSetSections.beforeFirst();
		
		while (rowSetSections.next()) {
			
			this.arrayListOfTotalQuantityOfItemsAddress.clear();
			this.arrayListOfTotalInvoiceAmountAddress.clear();
			
			/**
			 * Step 4.0: Zero out the section counters for hours and dollars.
			 */
			dtoInvoiceMetaData.setTotalHours(0);
			dtoInvoiceMetaData.setTotalInvoiceAmount(0);
			
			/*
			 * Step 4.1: Generate sheet for each sub-invoice.
			 */
			System.out.println("\nCreating spreadsheet tab for " + rowSetSections.getString("invoice_section_id") + ".");
			String stringInvoiceSectionID = rowSetSections.getString("invoice_section_id");
			Sheet sheet = workbook.createSheet(stringInvoiceSectionID);
			sheet.getPrintSetup().setLandscape(true);
			
			this.createSectionHeader(workbook, sheet, dtoInvoiceMetaData);

			/*
			 * Step 4.2: Generate header for details.
			 */
			this.createHeaderForDetails(sheet);

			/*
			 * Step 6: Get the invoice section details.
			 */
			intLoopIndex = sheet.getPhysicalNumberOfRows() + 1;
			RowSet rowSet = this.getDAOTimeCard().getInvoiceRowSet(this.getBillingPeriodID(), stringInvoiceSectionID);
			List<InvoiceLaborCategoryGroup> listOfInvoiceLaborCategoryGroup = this.getLaborCategoryGroups(rowSet);
			Row rowEmployeeDetail = null;
			
			
			int intEmployeeOrdinal = 0;
			
			ArrayList<String> arrayListOfQuantityAddress = new ArrayList<>();
			ArrayList<String> arrayListOfExtendedCostAddress = new ArrayList<>();
			
			for (InvoiceLaborCategoryGroup invoiceLaborCategoryGroup : listOfInvoiceLaborCategoryGroup) {
				
				Row rowDetail = sheet.createRow(intLoopIndex++);
				Cell cellLaborCategory = rowDetail.createCell(0);
				String stringLaborCategory = invoiceLaborCategoryGroup.getLaborCategoryID() + " - " + invoiceLaborCategoryGroup.getLaborCategoryName();
				cellLaborCategory.setCellValue(stringLaborCategory);
				cellLaborCategory.setCellStyle(cellStyleHeader);
	
				Cell cellEmployeeHeader = rowDetail.createCell(2);
				cellEmployeeHeader.setCellValue("Employee");
				cellEmployeeHeader.setCellStyle(cellStyleHeader);
				
				Cell cellLaborCategoryEmployeeCount = rowDetail.createCell(3);
				cellLaborCategoryEmployeeCount.setCellValue(invoiceLaborCategoryGroup.getEmployeeHoursSummary().size());
				cellLaborCategoryEmployeeCount.setCellStyle(cellStyleHeader);
				
				Cell cellQuantity = rowDetail.createCell(4);
				cellQuantity.setCellStyle(cellStyleExtendedCost);
				this.getQuantityOfItemsAddressList().add(cellQuantity.getAddress().formatAsString());
				dtoInvoiceMetaData.setTotalHours(dtoInvoiceMetaData.getTotalHours() + invoiceLaborCategoryGroup.getTotalHours());
				
				Cell cellUnitPrice = rowDetail.createCell(5);
				cellUnitPrice.setCellValue("");
				
				Cell cellExtendedAmount = rowDetail.createCell(6);
				cellExtendedAmount.setCellStyle(cellStyleExtendedCost);
				this.getTotalInvoiceAmountAddressList().add(cellExtendedAmount.getAddress().formatAsString());
				dtoInvoiceMetaData.setTotalInvoiceAmount(dtoInvoiceMetaData.getTotalInvoiceAmount() + invoiceLaborCategoryGroup.getTotalExtendedAmount());
				
				
				arrayListOfQuantityAddress.clear();
				arrayListOfExtendedCostAddress.clear();
				
				String stringExtendedCostCellFormula = "";
				
				for (EmployeeHoursSummary employeeHoursSummary : invoiceLaborCategoryGroup.getEmployeeHoursSummary()) {
					
					rowEmployeeDetail = sheet.createRow(intLoopIndex++);
					Cell cellBlank = rowEmployeeDetail.createCell(0);
					cellBlank.setCellValue("");

					intEmployeeOrdinal++;
					Cell cellEmployeeOrdinal = rowEmployeeDetail.createCell(1);
					cellEmployeeOrdinal.setCellValue(intEmployeeOrdinal);
					cellEmployeeOrdinal.setCellStyle(cellStyle);
					
					Cell cellEmployeeName = rowEmployeeDetail.createCell(2);
					cellEmployeeName.setCellValue(employeeHoursSummary.getLastName().trim() + ", " + employeeHoursSummary.getFirstName().trim() + " (" + employeeHoursSummary.getEmployeeID() + ")");
					cellEmployeeName.setCellStyle(cellStyle);
					
					cellBlank = rowEmployeeDetail.createCell(3);
					cellBlank.setCellValue("");

					Cell cellEmployeeHours = rowEmployeeDetail.createCell(4);
					cellEmployeeHours.setCellStyle(cellStyleHours);
					cellEmployeeHours.setCellValue(employeeHoursSummary.getProfitHours());
					arrayListOfQuantityAddress.add(cellEmployeeHours.getAddress().formatAsString());

					Cell cellEmployeeBillRate = rowEmployeeDetail.createCell(5);
					cellEmployeeBillRate.setCellStyle(cellStyleRate);
					cellEmployeeBillRate.setCellValue(employeeHoursSummary.getProfitHourlyRate());

					Cell cellExtendedCost = rowEmployeeDetail.createCell(6);
					stringExtendedCostCellFormula = "PRODUCT(" + cellEmployeeHours.getAddress().formatAsString() + "," 
													+ cellEmployeeBillRate.getAddress().formatAsString() + ")";
					cellExtendedCost.setCellStyle(cellStyleExtendedCost);
					cellExtendedCost.setCellFormula(stringExtendedCostCellFormula);
					/*cellExtendedCost.setCellValue(
							this.getTruncatedDouble(
									"########0.0000",
									employeeHoursSummary.getProfitHours() * employeeHoursSummary.getProfitHourlyRate()
								)
						);*/
					arrayListOfExtendedCostAddress.add(cellExtendedCost.getAddress().formatAsString());

					rowEmployeeDetail = sheet.createRow(intLoopIndex++);
					cellBlank = rowEmployeeDetail.createCell(0);
					cellBlank.setCellValue("");

					Cell cellEmployeeNonProfitOrdinal = rowEmployeeDetail.createCell(1);
					cellEmployeeNonProfitOrdinal.setCellValue(intEmployeeOrdinal);
					cellEmployeeNonProfitOrdinal.setCellStyle(cellStyleNonProfitEmployeeName);
					
					cellEmployeeName = rowEmployeeDetail.createCell(2);
					cellEmployeeName.setCellValue(
							employeeHoursSummary.getLastName().trim() + ", " + employeeHoursSummary.getFirstName().trim() + " (" + employeeHoursSummary.getEmployeeID() + ")" +
							" Government Billable Hours Minus Profit"
						);
					cellEmployeeName.setCellStyle(cellStyleNonProfitEmployeeName);
					sheet.addMergedRegion(new CellRangeAddress(intLoopIndex - 1,intLoopIndex - 1,2,3));

					cellEmployeeHours = rowEmployeeDetail.createCell(4);
					cellEmployeeHours.setCellValue(employeeHoursSummary.getNonProfitHours());
					cellEmployeeHours.setCellStyle(cellStyleNonProfitHours);
					arrayListOfQuantityAddress.add(cellEmployeeHours.getAddress().formatAsString());
					
					cellEmployeeBillRate = rowEmployeeDetail.createCell(5);
					cellEmployeeBillRate.setCellValue(employeeHoursSummary.getNonProfitHourlyRate());
					cellEmployeeBillRate.setCellStyle(cellStyleNonProfitRate);

					cellExtendedCost = rowEmployeeDetail.createCell(6);
					stringExtendedCostCellFormula = "PRODUCT(" + cellEmployeeHours.getAddress().formatAsString() 
													+ "," + cellEmployeeBillRate.getAddress().formatAsString() + ")";
					cellExtendedCost.setCellFormula(stringExtendedCostCellFormula);
					/*cellExtendedCost.setCellValue(
							this.getTruncatedDouble(
									"########0.0000",
									employeeHoursSummary.getNonProfitHours() * employeeHoursSummary.getNonProfitHourlyRate()
								)
						);*/
					cellExtendedCost.setCellStyle(cellStyleNonProfitExtendedCost);
					arrayListOfExtendedCostAddress.add(cellExtendedCost.getAddress().formatAsString());
				}
				
				/*
				 * Create the cell formula that calculates total hours for the CLIN.
				 */
				StringWriter stringWriterTotalHoursRange = new StringWriter();
				PrintWriter printWriterTotalHoursRange = new PrintWriter(stringWriterTotalHoursRange);
				
				printWriterTotalHoursRange.write("SUM(");
				
				for (int intIndex = 0; intIndex < arrayListOfQuantityAddress.size(); intIndex++) {
					
					printWriterTotalHoursRange.write(arrayListOfQuantityAddress.get(intIndex));
					
					if (intIndex != arrayListOfQuantityAddress.size() - 1) {
						printWriterTotalHoursRange.write(",");
					}
				}
				
				printWriterTotalHoursRange.write(")");
				printWriterTotalHoursRange.flush();
				printWriterTotalHoursRange.close();
				String stringCLINTotalHours = stringWriterTotalHoursRange.getBuffer().toString();
				
				cellQuantity.setCellFormula(stringCLINTotalHours);
				
				/*
				 * Create the cell formula that calculates total extended amount for the CLIN.
				 */
				StringWriter stringWriterExtendedAmountRange = new StringWriter();
				PrintWriter printWriterExtendedAmountRange = new PrintWriter(stringWriterExtendedAmountRange);
				
				printWriterExtendedAmountRange.write("SUM(");
				
				for (int intIndex = 0; intIndex < arrayListOfExtendedCostAddress.size(); intIndex++) {
					
					printWriterExtendedAmountRange.write(arrayListOfExtendedCostAddress.get(intIndex));
					
					if (intIndex != arrayListOfExtendedCostAddress.size() - 1) {
						printWriterExtendedAmountRange.write(",");
					}
				}
				
				printWriterExtendedAmountRange.write(")");
				printWriterExtendedAmountRange.flush();
				printWriterExtendedAmountRange.close();
				String stringCLINExtendedAmount = stringWriterExtendedAmountRange.getBuffer().toString();
				
				cellExtendedAmount.setCellFormula(stringCLINExtendedAmount);
			}
			
			for (int intColumnIndex = 0; intColumnIndex < rowEmployeeDetail.getLastCellNum(); intColumnIndex++) {
				sheet.autoSizeColumn(intColumnIndex);
			}
			
			Row rowInvoiceAggregates = sheet.createRow(intLoopIndex++);
			
			/*
			 * Create the cell formula that calculates the aggregate value for
			 * the total quantity of items.
			 */
			StringWriter stringWriterQuantityRange = new StringWriter();
			PrintWriter printWriterQuantityRange = new PrintWriter(stringWriterQuantityRange);
			
			printWriterQuantityRange.write("SUM(");
			
			for (int intIndex = 0; intIndex < this.getQuantityOfItemsAddressList().size(); intIndex++) {
				
				printWriterQuantityRange.write(this.getQuantityOfItemsAddressList().get(intIndex));
				
				if (intIndex != this.getQuantityOfItemsAddressList().size() - 1) {
					printWriterQuantityRange.write(",");
				}
			}
			
			printWriterQuantityRange.write(")");
			printWriterQuantityRange.flush();
			printWriterQuantityRange.close();
			String stringTotalQuantityOfItemsFormula = stringWriterQuantityRange.getBuffer().toString();
			
			/*
			 * Set the total quantity of items in the invoice header.
			 */
			sheet.getRow(5).getCell(6).setCellStyle(cellStyleTotalInvoiceHours);
			sheet.getRow(5).getCell(6).setCellFormula(stringTotalQuantityOfItemsFormula);
			
			/*
			 * Set the total quantity of items after the last row of data.
			 */
			Cell cellTotalQuantityOfItemsLabel = rowInvoiceAggregates.createCell(3);
			cellTotalQuantityOfItemsLabel.setCellStyle(cellStyleTotalInvoiceHours);
			cellTotalQuantityOfItemsLabel.setCellValue("Total Quantity of Items:");
			Cell cellTotalQuantityOfItems = rowInvoiceAggregates.createCell(4);
			cellTotalQuantityOfItems.setCellStyle(cellStyleTotalInvoiceHours);
			cellTotalQuantityOfItems.setCellFormula(stringTotalQuantityOfItemsFormula);
			
			/*
			 * Create the cell formula that calculates the aggregate value for
			 * the total invoice amount.
			 */
			StringWriter stringWriterTotalAmountRange = new StringWriter();
			PrintWriter printWriterTotalAmountRange = new PrintWriter(stringWriterTotalAmountRange);
			
			printWriterTotalAmountRange.write("SUM(");
			
			for (int intIndex = 0; intIndex < this.getTotalInvoiceAmountAddressList().size(); intIndex++) {
				
				printWriterTotalAmountRange.write(this.getTotalInvoiceAmountAddressList().get(intIndex));
				
				if (intIndex != this.getTotalInvoiceAmountAddressList().size() - 1) {
					printWriterTotalAmountRange.write(",");
				}
			}
			
			printWriterTotalAmountRange.write(")");
			printWriterTotalAmountRange.flush();
			printWriterTotalAmountRange.close();
			String stringTotalInvoiceAmountFormula = stringWriterTotalAmountRange.getBuffer().toString();
			
			/*
			 * Set the total invoice amount in the invoice header.
			 */
			sheet.getRow(6).getCell(6).setCellStyle(cellStyleTotalInvoiceAmount);
			sheet.getRow(6).getCell(6).setCellFormula(stringTotalInvoiceAmountFormula);
			
			/*
			 * Set the total invoice amount after the last row of data.
			 */
			Cell cellTotalInvoiceAmountLabel = rowInvoiceAggregates.createCell(5);
			cellTotalInvoiceAmountLabel.setCellStyle(cellStyleTotalInvoiceAmount);
			cellTotalInvoiceAmountLabel.setCellValue("Total Invoice Amount:");
			Cell cellTotalInvoiceAmount = rowInvoiceAggregates.createCell(6);
			cellTotalInvoiceAmount.setCellStyle(cellStyleTotalInvoiceAmount);
			cellTotalInvoiceAmount.setCellFormula(stringTotalInvoiceAmountFormula);
			
			/*
			 * Step 7: Generate the invoice footer
			 */
			this.createSectionFooter(sheet, dtoInvoiceMetaData);
		}
		
		/*
		 * Step 8: Write the invoice to a file.
		 */
		ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
		workbook.write(byteArrayOutputStream);
		byte[] arrayOfByte = byteArrayOutputStream.toByteArray();

		return arrayOfByte;
	}
	
	public static void main(String[] arrayOfStringArgument) throws Exception {		
		
		Scanner scannerForKeyboard = new Scanner(System.in);
		System.out.print("Enter billing period (yyyy.mm): ");
		String stringBillingPeriodIDNew = scannerForKeyboard.next();
		if (stringBillingPeriodIDNew == null || stringBillingPeriodIDNew.trim().length() == 0) {
			System.out.println("The billing period was not entered. Exiting program.");
			return;
		}
		
		InvoiceGenerator invoiceGenerator = new InvoiceGenerator();
		invoiceGenerator.setBillingPeriodID(stringBillingPeriodIDNew);
		
		byte[] arrayOfByte = invoiceGenerator.generateInvoice();
		
		String stringFileName = stringBillingPeriodIDNew + ".invoice.xlsx";
		FileOutputStream fileOutputStream = new FileOutputStream(new File(stringFileName));
		fileOutputStream.write(arrayOfByte);
		fileOutputStream.close();
		
		scannerForKeyboard.close();
	}
}