package com.bif.timecard.loader;

import java.io.BufferedReader;
import java.io.Console;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.text.DateFormatter;

import com.bif.timecard.loader.enumeration.CMTTokenIndex;
import com.bif.timecard.loader.enumeration.ConfigurationProperty;
import com.bif.timecard.loader.enumeration.PayType;
import com.bif.timecard.loader.enumeration.ReaderMode;
import com.bif.timecard.loader.enumeration.TokenIndex;

/**
 * <p>Instances of this class are used to load time card data into the BrainJack&trade; Time Card
 * repository.</p>
 * 
 * @author <a href="mailto:roderick.barnes@biftechnologies.com">Roderick L. Barnes, Sr.</a>
 */
public class TimeCardDataLoader {
	/**
	 * class object used to log events as they are generated
	 */
	private static Logger logger = Logger.getLogger(TimeCardDataLoader.class.getName());

	/**
	 * Member attribute providing a reference to the file from which data will be loaded.
	 */
	private File fileForData = null;
	
	/**
	 * Member attribute for the reader that will be used to load the data one line at a time
	 * from the data file.
	 */
	private BufferedReader bufferedReader = null;
	
	/**
	 * Collection of records loaded from the data file.
	 */
	private List<TimeCardRecord> listOfTimeCardRecord = new ArrayList<TimeCardRecord>();
	
	/**
	 * State variable used to keep track of the state that the loader is in. When a column header record is encountered
	 * it is ignored. Other lines are treated as data and loaded into the record collection {@code listOfTimeCardRecord}.
	 */
	private ReaderMode readerMode = ReaderMode.ColumnHeader;
	
	/**
	 * Returns a reference to the class logger. Call this method when using the logger in an instance context.
	 * @return a reference to the class logger.
	 */
	public Logger getLogger() {
		return TimeCardDataLoader.logger;
	}
	
	/**
	 * Returns a reference to the class logger. Call this method when using the logger in a static context.
	 * @return a reference to the class logger.
	 */
	public static Logger getClassLogger() {
		return TimeCardDataLoader.logger;
	}

	/**
	 * Returns a reference to the collection of time card records that were loaded from
	 * the data file.
	 * @return a reference to the collection of time card records loaded from the
	 * data file.
	 */
	public List<TimeCardRecord> getTimeCardRecords() {
		return this.listOfTimeCardRecord;
	}

	/**
	 * Sets the {@code BufferedReader} that will be used to load data from the data file.
	 * @param bufferedReaderNew the reader for reading the data file line by line.
	 */
	private void setReader(BufferedReader bufferedReaderNew) {
		this.bufferedReader = bufferedReaderNew;
	}
	
	/**
	 * Returns a reference to the reader for the data file.
	 * @return a reference to the reader for the data file.
	 */
	public BufferedReader getReader() {
		return this.bufferedReader;
	}
	
	/**
	 * Sets a reference to the file object that will be used to access the
	 * data file of time card records.
	 * @param fileForDataNew reference to the data file containing the records.
	 * @throws Exception thrown if the file does not exist.
	 */
	public void setFile(File fileForDataNew) throws Exception {
		// Step 1: Verify that the file exists.
		if (fileForDataNew.exists() == false) {
			throw new Exception("File does not exist: " + fileForDataNew.getName());
		}
		
		// Step 2: Assign the parameter to the class instance attribute.
		this.fileForData = fileForDataNew;
	}
	
	/**
	 * Returns a reference to the data file for the time card data.
	 * @return a reference to the data file for the time card data.
	 */
	public File getFile() {
		return this.fileForData;
	}
	
	/**
	 * Sets the reader mode for the time card loader. This is a state variable.
	 * @param readerModeNew the reader mode for the loader process.
	 */
	public void setReaderMode(ReaderMode readerModeNew) {
		this.readerMode = readerModeNew;
	}
	
	/**
	 * Returns the reader mode for the time card loader process.
	 * @return the reader mode for the time card loader process.
	 */
	public ReaderMode getReaderMode() {
		return this.readerMode;
	}
	
	/**
	 * Returns an enumeation entry indicating the reader mode that the loader should be in.
	 * The mode is based on the tokens found in the {@code String} array.
	 * @param arrayOfStringLineToken array of tokens read from a line i the data file.
	 * @return an enumeration entry indicating the reader mode of the loader. 
	 */
	public ReaderMode determineReaderMode(String[] arrayOfStringLineToken) {
		/**
		 * Step 1: Default to DataLine more.
		 */
		ReaderMode readerModeTemp = ReaderMode.DataLine;
				
		// Step 2: Test to see if we are in ColumnHeader mode or DataLine mode.
		// If it is a ColumnHeader line the first array element will be "Employee Identifier"
		// and the second element will be "Last Name".
		switch (ConfigurationProperty.CompanyID.getValue().toLowerCase().trim()) {
			case "xsi":
				if (arrayOfStringLineToken[TokenIndex.EmployeeIdentifier.ordinal()] != null && arrayOfStringLineToken[TokenIndex.EmployeeIdentifier.ordinal()].trim().equalsIgnoreCase("Employee Identifier")) {
					readerModeTemp = ReaderMode.ColumnHeader;
				}
				break;
			case "appleone":
				if (arrayOfStringLineToken[TokenIndex.EmployeeIdentifier.ordinal()] != null && arrayOfStringLineToken[TokenIndex.EmployeeIdentifier.ordinal()].trim().equalsIgnoreCase("Employee Identifier")) {
					readerModeTemp = ReaderMode.ColumnHeader;
				}
				break;
			case "cmt":
				if (arrayOfStringLineToken[CMTTokenIndex.EmployeeIdentifier.ordinal()] != null && arrayOfStringLineToken[CMTTokenIndex.EmployeeIdentifier.ordinal()].trim().equalsIgnoreCase("Employee ID")) {
					readerModeTemp = ReaderMode.ColumnHeader;
				}
				break;
		}
		
		// Step 3: Return the mode.
		return readerModeTemp;
	}
	
	/**
	 * Handles the conversion of a data file token into a numeric
	 * value suitable for a time card object ({@code TimeCardRecord}).
	 * @param stringPotentialNumeric value that may be numeric.
	 * @return Double object corresponding to the numeric value of the 
	 * 		{@code String} that is passed. If a {@code null}, dash, or empty
	 * 		string is passed it will be converted into zero.
	 */
	private Double getNumericValue(String stringPotentialNumeric) {
		if (stringPotentialNumeric == null || stringPotentialNumeric.trim().length() == 0 || stringPotentialNumeric.trim().equalsIgnoreCase("-")) {
			return 0.0;
		} else {
			return Double.parseDouble(stringPotentialNumeric);
		}
	}

	/**
	 * Returns a normalized labor category name.
	 * @param stringLaborCategoryName a normalized labor category name.
	 * @return a normalized labor category name.
	 */
	private String getNormalizedLaborCategoryID(String stringLaborCategoryID) {
		/**
		 * Make the labor category name upper case.
		 */
		if (ConfigurationProperty.CompanyID.getValue().equalsIgnoreCase("xsi")) {
			stringLaborCategoryID = stringLaborCategoryID.replace("001", "0001");
		}
		
		if (ConfigurationProperty.CompanyID.getValue().equalsIgnoreCase("cmt")) {
			stringLaborCategoryID = stringLaborCategoryID.replace("001", "0001");
		}
				
		return stringLaborCategoryID;
	}
	
	/**
	 * Returns a pay type name without special characters.
	 * @param stringEventName event type name as it appears in the data file.
	 * @return pay type name without special characters.
	 */
	private String getNormalizedPayTypeName(String stringEventName) {
		if (stringEventName.equalsIgnoreCase("Regular Time")) {
			stringEventName = "Work";
		} else if (stringEventName.equalsIgnoreCase("Scheduled Vacation")) {
			stringEventName = "Vacation";
		} else if (stringEventName.equalsIgnoreCase("Unscheduled Sick")) {
			stringEventName = "Sick";
		} else if (stringEventName.equalsIgnoreCase("FMLA - Unscheduled Sick")) {
			stringEventName = "Sick";
		} else if (stringEventName.equalsIgnoreCase("FMLA - Scheduled Vacation")) {
			stringEventName = "Vacation";
		}
		
		stringEventName = stringEventName.replace("Work/Weather", "Work Weather");
		return stringEventName.replace("\\", " ");
	}
	
	/**
	 * Returns a {@code TimeCardRecord} object based on the tokens in the
	 * array {@arrayOfStringLineToken}. The XSI file format will be used to
	 * map tokens to object attributes.
	 * @param arrayOfStringLineToken array of tokens based on reading a line of data.
	 * @return a {@code TimeCardRecord} object
	 */
	public TimeCardRecord parseXSILine(String stringFileName, long longLineNumber, String stringLineOfText, String[] arrayOfStringLineToken) {
		String stringCompanyIDNew = ConfigurationProperty.CompanyID.getValue();
		String stringEmployeeIDNew = arrayOfStringLineToken[TokenIndex.EmployeeIdentifier.ordinal()];
		String stringFirstNameNew = arrayOfStringLineToken[TokenIndex.FirstName.ordinal()];
		String stringLastNameNew = arrayOfStringLineToken[TokenIndex.LastName.ordinal()];
		String stringEventDateNew = arrayOfStringLineToken[TokenIndex.EventDate.ordinal()];
		String stringPayTypeNew = arrayOfStringLineToken[TokenIndex.PayType.ordinal()];
		stringPayTypeNew = this.getNormalizedPayTypeName(stringPayTypeNew);
		PayType payTypeNew = PayType.captionToPayType(stringPayTypeNew);

		String stringRegularDurationNew = arrayOfStringLineToken[TokenIndex.RegularDuration.ordinal()];
		Double doubleRegularDurationNew = this.getNumericValue(stringRegularDurationNew);
		
		String stringUnpaidDurationNew = arrayOfStringLineToken[TokenIndex.UnpaidDuration.ordinal()];
		Double doubleUnpaidDurationNew = this.getNumericValue(stringUnpaidDurationNew);

		String stringTotalPaidDurationNew = arrayOfStringLineToken[TokenIndex.TotalPaidDuration.ordinal()];
		Double doubleTotalPaidDurationNew = Double.parseDouble(stringTotalPaidDurationNew);

		String stringProjectIDNew = arrayOfStringLineToken[TokenIndex.ProjectID.ordinal()];
		String stringProjectNameNew = arrayOfStringLineToken[TokenIndex.ProjectName.ordinal()];
		String stringLaborCategoryIDNew = arrayOfStringLineToken[TokenIndex.LaborCategoryID.ordinal()];
		stringLaborCategoryIDNew = this.getNormalizedLaborCategoryID(stringLaborCategoryIDNew);
		
		String stringDepartmentIDNew = arrayOfStringLineToken[TokenIndex.DepartmentID.ordinal()];
		String stringDepartmentNameNew = arrayOfStringLineToken[TokenIndex.DepartmentName.ordinal()];
		
		Double doubleHourlyRateNew = 10.0;
		
		TimeCardRecord timeCardRecord = new TimeCardRecord(
				stringFileName,
				longLineNumber,
				stringLineOfText,
				stringCompanyIDNew,
				stringEmployeeIDNew, 
				stringFirstNameNew,
				stringLastNameNew,
				stringEventDateNew,
				payTypeNew,
				doubleRegularDurationNew,
				doubleUnpaidDurationNew,
				doubleTotalPaidDurationNew,
				stringProjectIDNew,
				stringProjectNameNew,
				stringLaborCategoryIDNew,
				stringDepartmentIDNew,
				stringDepartmentNameNew,
				doubleHourlyRateNew
			);

		return timeCardRecord;

	}
	
	/**
	 * Returns a {@code TimeCardRecord} object based on the tokens in the
	 * array {@arrayOfStringLineToken}. The AppleOne file format will be used to
	 * map tokens to object attributes.
	 * @param arrayOfStringLineToken array of tokens based on reading a line of data.
	 * @return a {@code TimeCardRecord} object
	 */
	public TimeCardRecord parseAppleOneLine(String stringFileName, long longLineNumber, String stringLineOfText, String[] arrayOfStringLineToken) {
		String stringCompanyIDNew = ConfigurationProperty.CompanyID.getValue();
		String stringEmployeeIDNew = arrayOfStringLineToken[TokenIndex.EmployeeIdentifier.ordinal()];
		
		/**
		 * In the AppleOne data sets the first name and last name are in one field separated
		 * by a comma.
		 */
		String stringFirstNameNew = arrayOfStringLineToken[TokenIndex.LastName.ordinal()];
		stringFirstNameNew = stringFirstNameNew.replace("\"", "").trim();
		String stringLastNameNew = arrayOfStringLineToken[TokenIndex.FirstName.ordinal()];
		stringLastNameNew = stringLastNameNew.replace("\"", "").trim();
		
		String stringEventDateNew = arrayOfStringLineToken[TokenIndex.EventDate.ordinal()];
		String stringPayTypeNew = arrayOfStringLineToken[TokenIndex.PayType.ordinal()];
		stringPayTypeNew = this.getNormalizedPayTypeName(stringPayTypeNew);

		PayType payTypeNew = PayType.captionToPayType(stringPayTypeNew);
		if (payTypeNew == null) {
			System.out.println("stringPayTypeNew=\"" + stringPayTypeNew  + "\"");
		}

		String stringRegularDurationNew = arrayOfStringLineToken[TokenIndex.RegularDuration.ordinal()];
		Double doubleRegularDurationNew = this.getNumericValue(stringRegularDurationNew);
		
		String stringUnpaidDurationNew = arrayOfStringLineToken[TokenIndex.UnpaidDuration.ordinal()];
		Double doubleUnpaidDurationNew = this.getNumericValue(stringUnpaidDurationNew);

		String stringTotalPaidDurationNew = arrayOfStringLineToken[TokenIndex.TotalPaidDuration.ordinal()];
		Double doubleTotalPaidDurationNew = this.getNumericValue(stringTotalPaidDurationNew);

		String stringProjectIDNew = arrayOfStringLineToken[TokenIndex.ProjectID.ordinal()].trim();
		String stringProjectNameNew = arrayOfStringLineToken[TokenIndex.ProjectName.ordinal()].trim();
		
		/**
		 * AppleOne labor category identifiers have an extra leading zero.
		 */
		String stringLaborCategoryIDNew = arrayOfStringLineToken[TokenIndex.LaborCategoryID.ordinal()];
		stringLaborCategoryIDNew = getNormalizedLaborCategoryID(stringLaborCategoryIDNew);
				
		String stringDepartmentIDNew = arrayOfStringLineToken[TokenIndex.DepartmentID.ordinal()];
		String stringDepartmentNameNew = arrayOfStringLineToken[TokenIndex.DepartmentName.ordinal()];
		
		Double doubleHourlyRateNew = 10.0;
		
		TimeCardRecord timeCardRecord = new TimeCardRecord(
				stringFileName,
				longLineNumber,
				stringLineOfText,
				stringCompanyIDNew,
				stringEmployeeIDNew, 
				stringFirstNameNew,
				stringLastNameNew,
				stringEventDateNew,
				payTypeNew,
				doubleRegularDurationNew,
				doubleUnpaidDurationNew,
				doubleTotalPaidDurationNew,
				stringProjectIDNew,
				stringProjectNameNew,
				stringLaborCategoryIDNew,
				stringDepartmentIDNew,
				stringDepartmentNameNew,
				doubleHourlyRateNew
			);
		
		return timeCardRecord;		
	}

	/**
	 * Returns a {@code TimeCardRecord} object based on the tokens in the
	 * array {@arrayOfStringLineToken}. The CMT (Consulting, Management, and Technologyh)
	 * file format will be used to map tokens to object attributes.
	 * @param arrayOfStringLineToken array of tokens based on reading a line of data.
	 * @return a {@code TimeCardRecord} object
	 */
	public TimeCardRecord parseCMTLine(String stringFileName, long longLineNumber, String stringLineOfText, String[] arrayOfStringLineToken) throws Exception {
		String stringCompanyIDNew = ConfigurationProperty.CompanyID.getValue();
		String stringEmployeeIDNew = arrayOfStringLineToken[CMTTokenIndex.EmployeeIdentifier.ordinal()];
		
		/**
		 * In the CMT data sets the first name and last name are in one field separated
		 * by a comma.
		 */
		String stringLastNameNew = arrayOfStringLineToken[CMTTokenIndex.LastName.ordinal()].trim();
		stringLastNameNew = stringLastNameNew.replace("\"", "").trim();
		String stringFirstNameNew = arrayOfStringLineToken[CMTTokenIndex.FirstName.ordinal()].trim();
		stringFirstNameNew = stringFirstNameNew.replace("\"", "").trim();
		
		String stringEventDurationNew = arrayOfStringLineToken[CMTTokenIndex.EventDuration.ordinal()].trim();
		Double doubleEventDurationNew = this.getNumericValue(stringEventDurationNew);
		

		String stringEventDateNew = arrayOfStringLineToken[CMTTokenIndex.EventDate.ordinal()].trim();
		SimpleDateFormat simpleDateFormat = new SimpleDateFormat("MM/dd/yy");
		Date dateEventNew = simpleDateFormat.parse(stringEventDateNew);
		SimpleDateFormat simpleDateFormatFourDigitYear = new SimpleDateFormat("MM/dd/yyyy");
		stringEventDateNew = simpleDateFormatFourDigitYear.format(dateEventNew);

		String stringEventDayOfWeek = arrayOfStringLineToken[CMTTokenIndex.EventDayOfWeek.ordinal()].trim();
		
		String stringPayTypeNew = arrayOfStringLineToken[CMTTokenIndex.EventType.ordinal()];
		stringPayTypeNew = this.getNormalizedPayTypeName(stringPayTypeNew);

		PayType payTypeNew = PayType.captionToPayType(stringPayTypeNew);
		if (payTypeNew == null) {
			System.out.println("stringPayTypeNew=\"" + stringPayTypeNew  + "\"");
		}

		String stringHours = arrayOfStringLineToken[CMTTokenIndex.EventDuration.ordinal()].trim();
		Double doubleHours = this.getNumericValue(stringHours);
		
		
		String stringRegularDurationNew = "0.0"; //arrayOfStringLineToken[CMTTokenIndex.EventDuration.ordinal()];
		Double doubleRegularDurationNew = this.getNumericValue(stringRegularDurationNew);
			
		String stringUnpaidDurationNew = "0.0"; //arrayOfStringLineToken[TokenIndex.UnpaidDuration.ordinal()];
		Double doubleUnpaidDurationNew = this.getNumericValue(stringUnpaidDurationNew);

		String stringTotalPaidDurationNew = "0.0"; //arrayOfStringLineToken[TokenIndex.TotalPaidDuration.ordinal()];
		Double doubleTotalPaidDurationNew = this.getNumericValue(stringTotalPaidDurationNew);

		switch (payTypeNew) {
			case Bereavement:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case Break:
				doubleRegularDurationNew = 0.0;
				doubleUnpaidDurationNew = doubleHours;
				doubleTotalPaidDurationNew = 0.0;
				break;			
			case Holiday:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case JuryDuty:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case Meal:
				doubleRegularDurationNew = 0.0;
				doubleUnpaidDurationNew = doubleHours;
				doubleTotalPaidDurationNew = 0.0;
				break;
			case MilitaryLeave:
				doubleRegularDurationNew = 0.0;
				doubleUnpaidDurationNew = doubleHours;
				doubleTotalPaidDurationNew = 0.0;
				break;
			case Sick:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case UnpaidTime:
				doubleRegularDurationNew = 0.0;
				doubleUnpaidDurationNew = doubleHours;
				doubleTotalPaidDurationNew = 0.0;
				break;
			case Vacation:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case WeatherGovernmentClosure:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case WorkWeatherClosure:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
			case Work:
				doubleRegularDurationNew = doubleHours;
				doubleUnpaidDurationNew = 0.0;
				doubleTotalPaidDurationNew = doubleHours;
				break;
		}
		
		
		String stringProjectIDNew = arrayOfStringLineToken[CMTTokenIndex.ProjectID.ordinal()].trim();
		String stringProjectNameNew = arrayOfStringLineToken[CMTTokenIndex.ProjectName.ordinal()].trim();			
		
		/**
		 * CMT labor category identifiers have an extra leading zero.
		 */
		String stringLaborCategoryIDNew = arrayOfStringLineToken[CMTTokenIndex.LaborCategoryID.ordinal()];
		stringLaborCategoryIDNew = getNormalizedLaborCategoryID(stringLaborCategoryIDNew);
				
		String stringDepartmentIDNew = arrayOfStringLineToken[CMTTokenIndex.DepartmentID.ordinal()];
		String stringDepartmentNameNew = arrayOfStringLineToken[CMTTokenIndex.DepartmentName.ordinal()];
		
		Double doubleHourlyRateNew = 10.0;
		
		TimeCardRecord timeCardRecord = new TimeCardRecord(
				stringFileName,
				longLineNumber,
				stringLineOfText,
				stringCompanyIDNew,
				stringEmployeeIDNew, 
				stringFirstNameNew,
				stringLastNameNew,
				stringEventDateNew,
				payTypeNew,
				doubleRegularDurationNew,
				doubleUnpaidDurationNew,
				doubleTotalPaidDurationNew,
				stringProjectIDNew,
				stringProjectNameNew,
				stringLaborCategoryIDNew,
				stringDepartmentIDNew,
				stringDepartmentNameNew,
				doubleHourlyRateNew
			);
		
		return timeCardRecord;		
	}
	
	/**
	 * Parses lines from the time card data file.
	 * @param arrayOfStringLineToken array of {@code String} objects that comprise a line of
	 * data read from the time card data file.
	 * @return time card record
	 */
	public TimeCardRecord parseLine(String stringFileNameNew, long longLineNumberNew, String stringLineOfTextNew, String[] arrayOfStringLineToken) throws Exception {
		if (ConfigurationProperty.CompanyID.getValue().equalsIgnoreCase("xsi")) {
			return this.parseXSILine(stringFileNameNew, longLineNumberNew, stringLineOfTextNew, arrayOfStringLineToken);
		} else if (ConfigurationProperty.CompanyID.getValue().equalsIgnoreCase("appleone")) { 
			return this.parseAppleOneLine(stringFileNameNew, longLineNumberNew, stringLineOfTextNew, arrayOfStringLineToken);
		} else if (ConfigurationProperty.CompanyID.getValue().equalsIgnoreCase("cmt")) {
			return this.parseCMTLine(stringFileNameNew, longLineNumberNew, stringLineOfTextNew, arrayOfStringLineToken);
		}
		
		throw new Exception("Company ID is undefined.");
	}
	
	/**
	 * Returns a JDBC connection that will be used for creating time card records in the database.
	 * @return  a JDBC connection that will be used for creating time card records in the database.
	 * @throws SQLException thrown if there is a problem establishing the JDBC connection
	 * @throws ClassNotFoundException thrown if the JDBC driver class cannot be found on the classpath
	 */
	public Connection getConnection() throws SQLException, ClassNotFoundException {
		Connection connection = null;
		
		Class.forName(ConfigurationProperty.JDBCDriverClass.getValue());
		connection = DriverManager.getConnection(
				ConfigurationProperty.JDBCURL.getValue(),
				ConfigurationProperty.JDBCUsername.getValue(),
				ConfigurationProperty.JDBCPassword.getValue()
			);
		
		return connection;
	}
	
	/** 
	 * Returns a SQL resource (SELECT, INSERT, UPDATE, DELETE statement) based on the given
	 * resource name.
	 * @return a SQL resource that can be used in JDBC calls.
	 */
	public String getSQLResource(String stringResourceName) throws IOException {
		InputStream inputStream = this.getClass().getResourceAsStream(stringResourceName);
		InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
		BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
		
		String stringSQLResource = "";
		String stringLineOfText = null;
		while ((stringLineOfText = bufferedReader.readLine()) != null) {
			stringSQLResource += stringLineOfText + "\n";
		}
		
		return stringSQLResource;
	}
	
	/**
	 * Removes all records from the fact table {@code TIMECARD.TIME_CARD_FACT}.
	 * @return count of records deleted
	 */
	private long clearTimeCardRecords() {
		long longRecordsCleared = 0;
		Connection connection = null;
		PreparedStatement preparedStatement = null;

		try {
			/**
			 * Step 1: Get the SQL resource required for deleting the records.
			 */
			String stringSQL = this.getSQLResource("sql_resource_delete_time_card_data.sql");
			
			/**
			 * Step 2: Get JDBC connection required for dispatching the SQL command.
			 */
			connection = this.getConnection();
			
			/**
			 * Step 3: Create the statement object.
			 */
			preparedStatement = connection.prepareStatement(stringSQL);
			
			/**
			 * Step 4: Setting the company ID.
			 */
			preparedStatement.setString(1, ConfigurationProperty.CompanyID.getValue());
			this.getLogger().log(Level.INFO, "       Company ID: " + ConfigurationProperty.CompanyID.getValue());
			
			/**
			 * Step 5: Setting the billing period id.
			 */
			preparedStatement.setString(2, ConfigurationProperty.BillingPeriodID.getValue());
			this.getLogger().log(Level.INFO, "Billing Period ID: " + ConfigurationProperty.BillingPeriodID.getValue());
			
			/**
			 * Step 6: Delete the records.
			 */
			this.getLogger().log(Level.INFO, "Clearing time card records...");
			longRecordsCleared += preparedStatement.executeUpdate();
			this.getLogger().log(Level.INFO, "Records cleared: " + longRecordsCleared);
			
			/**
			 * Step 7: Close statement and connection objects.
			 */
			preparedStatement.close();
			preparedStatement = null;
			connection.close();
			connection = null;
		} catch (Exception exception) {
			exception.printStackTrace();
		} finally {
			if (preparedStatement != null) {
				try {
					preparedStatement.close();
				} catch (Exception exception) {
					exception.printStackTrace();
				}
			}

			if (connection != null) {
				try {
					connection.close();
				} catch (Exception exception) {
					exception.printStackTrace();
				}
			}			
		}
		
		return longRecordsCleared;
	}
	
	private long logRecordSaveFailure(Exception exceptionForFailure, Connection connection, TimeCardRecord timeCardRecord) {
		long longLogRecordsCreated = 0;
		PreparedStatement preparedStatement = null;
		try {
			String stringMessage = "Exception: " + exceptionForFailure.getMessage() + " - " + timeCardRecord.getFileName() + ":Line " + timeCardRecord.getLineNumber() + ":" + timeCardRecord.getLineOfText();
			this.getLogger().log(Level.SEVERE, stringMessage);
			
			preparedStatement = connection.prepareStatement("INSERT INTO timecard.log l (l.log_type, l.message) VALUES ('Exception',?)");
			preparedStatement.setString(1, stringMessage);
			preparedStatement.executeUpdate();
			preparedStatement.close();
			preparedStatement = null;
		} catch (Exception exception) {
			exception.printStackTrace();
		} finally {
			try {
				if (preparedStatement != null) {
					preparedStatement.close();
				}
			} catch (SQLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		return longLogRecordsCreated;
	}
	
	/**
	 * Persist the collected time card records ({@code TimeCardRecord}) to the
	 * time card repository.
	 */
	private long saveTimeCardRecords() {
		long longRecordsCreated = 0;
		Connection connection = null;
		PreparedStatement preparedStatement = null;
		double doublePercentComplete = 0;
		
		try {
			/**
			 * Step 1: Get the PL/SQL anonymous block that will be used to create the time card fact record.
			 */
			String stringCreateEmployeeTimeCardEntry = this.getSQLResource("sql_resource_create_time_card_fact.sql");

			/**
			 * Step 2: Get the JDBC connection to the time card database.
			 */
			connection = this.getConnection();
			
			/**
			 * Step 3: Create the time card records.
			 */
			preparedStatement = connection.prepareStatement(stringCreateEmployeeTimeCardEntry);
			
			this.getLogger().log(Level.INFO, "Saving time card records...");
			long longRecordsProcessed = 0;
			for (TimeCardRecord timeCardRecord : this.getTimeCardRecords()) {				
				/**
				 * If the billing period of the time card does not match the billing
				 * period of the data load skip it.
				 */
				if (timeCardRecord.getBillingPeriodID().equalsIgnoreCase(ConfigurationProperty.BillingPeriodID.getValue()) == false) {
					this.getLogger().log(
							Level.INFO, 
							"\n\tBilling period of time card record (" + timeCardRecord.getBillingPeriodID() + ") does not match data load (" + ConfigurationProperty.BillingPeriodID.getValue()+ ")." +
							"\n\tSkippng record for " + timeCardRecord.getLastName() + ", " + timeCardRecord.getFirstName() + 
							" (Employee ID: " + timeCardRecord.getEmployeeID() + ", Event Date: " + timeCardRecord.getEventDate()+ ")"
						);
					continue;
				}
				
				try {
					longRecordsProcessed++;
					preparedStatement.setString(1,	timeCardRecord.getCompanyID());
					preparedStatement.setString(2,	timeCardRecord.getEmployeeID());
					preparedStatement.setString(3,	timeCardRecord.getFirstName());
					preparedStatement.setString(4,	timeCardRecord.getLastName());
					preparedStatement.setString(5,	timeCardRecord.getEventDate());
					preparedStatement.setString(6,	timeCardRecord.getPayType().getCaption());
					preparedStatement.setDouble(7,	timeCardRecord.getRegularDuration());
					preparedStatement.setDouble(8,	timeCardRecord.getUnpaidDuration());
					preparedStatement.setDouble(9,	timeCardRecord.getTotalPaidDuration());
					preparedStatement.setString(10,	timeCardRecord.getProjectID());
					preparedStatement.setString(11,	timeCardRecord.getProjectName());
					preparedStatement.setString(12,	timeCardRecord.getLaborCategoryID());
					preparedStatement.setString(13,	timeCardRecord.getDepartmentID());
					preparedStatement.setString(14,	timeCardRecord.getDepartmentName());
					preparedStatement.setDouble(15,	timeCardRecord.getHourlyRate());
					longRecordsCreated += preparedStatement.executeUpdate();
				} catch (Exception exception) {
					// If the attempt to save the record fails create a log of the failure.
					this.logRecordSaveFailure(exception, connection, timeCardRecord);
				}
				
				doublePercentComplete = ((double)longRecordsCreated) / ((float)this.getTimeCardRecords().size()) * 100;
				
				if (longRecordsProcessed % 100 == 0) {
					this.getLogger().log(Level.INFO, String.format("Processed: %5d Records Loaded: %7d Complete: %5.2f", longRecordsProcessed, longRecordsCreated, doublePercentComplete));
				}
			}
			
			this.getLogger().log(Level.INFO, "Records created: " + longRecordsCreated);
			
			preparedStatement.close();
			preparedStatement = null;
			connection.close();
			connection = null;
		} catch (Exception exception) {
			exception.printStackTrace();
		} finally {
			if (preparedStatement != null) {
				try {
					preparedStatement.close();
				} catch (Exception exception) {
					exception.printStackTrace();
				}
			}

			if (connection != null) {
				try {
					connection.close();
				} catch (Exception exception) {
					exception.printStackTrace();
				}
			}			
		}
		
		return longRecordsCreated;
	}
	
	/**
	 * This method loads the time card data file into into a collection. The collection
	 * is then persisted to the time card repository.
	 * @throws IOException thrown if there is a problem accessing the time card data file.
	 */
	public void loadFile() throws IOException {
		TimeCardRecord timeCardRecordBetaTemp = null;

		// Step 0: Clear the collection of time card records from this instance.
		this.getTimeCardRecords().clear();

		// Step 1: Create objects required for reading the data file.
		FileReader fileReader = new FileReader(this.getFile());
		BufferedReader bufferedReader = new BufferedReader(fileReader);
		
		this.setReader(bufferedReader);
		
		// Step 2: Iterate through the lines in the file.
		long longLineNumber = 0;
		String stringLine = null;
		while ((stringLine = this.getReader().readLine()) != null) {
			try {
				// Step 2.1: Increment the line counter.
				longLineNumber++;
				
				// Step 2.2: Split the line of text into tokens.
				String[] arrayOfStringLineToken = stringLine.split(",");
				
				// Step 2.3: Determine what reading mode we are in.
				ReaderMode readerModeTemp = this.determineReaderMode(arrayOfStringLineToken);

				switch (readerModeTemp) {
					case ColumnHeader:
						timeCardRecordBetaTemp = null;
						break;
					case DataLine:
						timeCardRecordBetaTemp = this.parseLine(this.getFile().getName(), longLineNumber, stringLine, arrayOfStringLineToken);
						break;
				}

				if (timeCardRecordBetaTemp != null) { 
					this.getTimeCardRecords().add(timeCardRecordBetaTemp);
					if (longLineNumber % 500 == 0){
						this.getLogger().log(Level.INFO, String.format("Line %6d, Mode=%-20s: %-300s",longLineNumber, readerModeTemp, timeCardRecordBetaTemp));
					}
				}
			} catch (Exception exception) {
				// report exception
				exception.printStackTrace();
			}
		}
		
		this.clearTimeCardRecords();
		
		this.saveTimeCardRecords();
		
		try {
			this.getReader().close();
		} catch(Exception exception) {
			exception.printStackTrace();
		}
	}
	
	/**
	 * Constructor for new instances of the time card data loader. The provided
	 * file reference will be opened and read for time card records.
	 * @param fileForDataNew file containing time card data to be loaded into a repository
	 * @throws Exception thrown if there is a problem accessing the file, loading the data
	 * into a collection, or persisting the data to a database.
	 */
	public TimeCardDataLoader(File fileForDataNew) throws Exception {
		this.setFile(fileForDataNew);
	}
	
	/**
	 * Constructor for new instances of the time card data loader. The provided
	 * file matching the given will be opened and read for time card records.
	 * @param stringFileForDataNew name and path of the file containing time card 
	 * data to be loaded into a repository
	 * @throws Exception thrown if there is a problem accessing the file, loading the data
	 * into a collection, or persisting the data to a database.
	 */
	public TimeCardDataLoader(String stringFileForDataNew) throws Exception {
		this(new File(stringFileForDataNew));
	}
	
	private static String[] getHistory() {
		String[] arrayOfStringHistory = new String[]{
				"0.0.1 - Initial version. This version only loaded XSI data from a format that listed a week at a time on one line.",
				"0.1.0 - Made it possible to load data from a PayChex export. Only XSI data is loaded.",
				"0.4.0 - Made it possible to load data for AppleOne. Improved handling of labor category information.",
				"0.8.0 - Made it possible to load data for CMT. Improved error reporting.",
				"0.8.1 - Added a detailed debugging message for attempting to load data for an invalid labor category and billing period.",
				"1.1.5 - Simplified loading prompts. Made it possible to derive data file path. Deleting prior records automatically.",
				"1.5.0 - Made it possible to load data through the PayChex API using BrainJack Stargate.",
				"1.6.0 - Made it possible to load data through the QuickBooks API using BrainJack Stargate.",
				"1.7.0 - Made it possible to load data through the APD API using BrainJack Stargate."
			};
		
		return arrayOfStringHistory;
	}
	
	private static void displayLoaderInfo() {
		System.out.println("BrainJack Time Card Data Loader");
		System.out.println("ver 1.1.5, Copyright 2019");
		System.out.println("-------------------------------");
	}
	
	/**
	 * Execution harness for the time card data loader. Parameters that are found in the
	 * configuration file will be used to run the loader.
	 * @param arrayOfStringParameter one parameter specifying the location of the
	 * configuration file is provided on the command line.
	 * @throws Exception thrown if there is a problem opening the configuration file,
	 * opening the data file, loading the data into a collection, or persisting the data
	 * to a database. 
	 */
	public static void main(String[] arrayOfStringParameter) throws Exception {
		/**
		 * Display the data loader banner, version number, and copyright.
		 */
		TimeCardDataLoader.displayLoaderInfo();
		
		/**
		 * Step 1: Declare variables need for the launch harness.
		 */
		String stringConfigurationFileLocation = null;
		String stringValueNew = null;
		
		/**
		 * Step 2: Extract the name of the configuration file from the parameters.
		 */
		if (arrayOfStringParameter.length > 0) {
			stringConfigurationFileLocation = arrayOfStringParameter[0];
		}
		
		/**
		 * Step 2.1a: If the configuration file was not provided 
		 * initiate manual prompt for the parameters.
		 */
		if (stringConfigurationFileLocation == null) {
			Scanner scanner = new Scanner(System.in);
			for (ConfigurationProperty configurationProperty : ConfigurationProperty.values()) {
				/**
				 * Do not prompt for database URL, JDBC driver, database user, or password.
				 * Use the defaults in the software.
				 */
				switch (configurationProperty) {
					case JDBCURL:
						configurationProperty.setValue(configurationProperty.getDefault());
						continue;
					case JDBCDriverClass:
						configurationProperty.setValue(configurationProperty.getDefault());
						continue;
					case JDBCUsername:
						configurationProperty.setValue(configurationProperty.getDefault());
						continue;
					case JDBCPassword:
						configurationProperty.setValue(configurationProperty.getDefault());
						continue;
				}
				
				System.out.print(configurationProperty.getPrompt() + ": ");
				stringValueNew = scanner.nextLine().trim();
				
				/**
				 * If the user did not enter a value use the default value
				 * found in the enumeration entry.
				 */
				if (stringValueNew == null || stringValueNew.trim().length() == 0) {
					stringValueNew = configurationProperty.getDefault();
					System.out.println("\tUsing default: " + stringValueNew);
				}
				
				/**
				 * Assign the value to the enumeration entry.
				 */
				configurationProperty.setValue(stringValueNew);
			}
			scanner.close();
			
		/*
		 * Step 2.1b: If the configuration file was provided, load
		 * the properties from the provided file and prompt for
		 * any properties that are left undefined.
		 */
		} else {
			File file = new File(stringConfigurationFileLocation);
			Scanner scanner = new Scanner(file);
			
			String stringPropertyAssignment = null;
			String stringProperty = null;
			
			/*
			 * Store the properties identified as key-pair values.
			 */
			HashMap<String, String> hashMapOfConfigurationProperties = new HashMap<String, String>();
			
			/*
			 * Parse each line of the configuration file.
			 */
			while (scanner.hasNextLine()) {
				stringPropertyAssignment = scanner.nextLine();
				
				try {
					/*
					 * Extract and store the property definition as a
					 * key-pair value.
					 */
					stringProperty = stringPropertyAssignment.split("=")[0];
					stringValueNew = stringPropertyAssignment.split("=")[1];
					hashMapOfConfigurationProperties.put(stringProperty, stringValueNew);
					
				} catch (Exception exception) {
					TimeCardDataLoader.getClassLogger().log(Level.SEVERE, "Improperly formatted property declaration: " + stringPropertyAssignment);
					TimeCardDataLoader.getClassLogger().log(Level.SEVERE, "This line will not be processed.");
				}
			}
			scanner.close();
			
			/*
			 * Prepare the scanner for user input from the terminal.
			 */
			scanner = new Scanner(System.in);
			String stringConfigurationValue = null;
			
			/*
			 * Iterate through the expected properties to ensure that all required
			 * values have been provided.
			 */
			for (ConfigurationProperty configurationProperty : ConfigurationProperty.values()) {
				try {
					/*
					 * Determine whether the expected property was found in the user-specified
					 * configuration file.
					 */
					stringConfigurationValue = hashMapOfConfigurationProperties.get(configurationProperty.name());
					
				} catch (Exception exception) {
					stringConfigurationValue = null;
				}
				
				/*
				 * If a value for the expected property was not defined, prompt the
				 * user for a new value.
				 */
				if (stringConfigurationValue == null || stringConfigurationValue.length() == 0) {
					
					System.out.print(configurationProperty.getPrompt() + ": ");
					
					if (configurationProperty.equals(ConfigurationProperty.JDBCPassword)) {
						/*
						 * If the current property is a password, mask
						 * the user input in the console.
						 */
						Console console = System.console();
						stringConfigurationValue = String.valueOf(console.readPassword());
					} else {
						stringConfigurationValue = scanner.nextLine().trim();
					}
					
					/*
					 * If the user does not explicitly define a value when
					 * prompted, assign the property default value.
					 */
					if (stringConfigurationValue == null || stringConfigurationValue.trim().length() == 0) {
						stringConfigurationValue = configurationProperty.getDefault();
						System.out.println("\tUsing default: " + stringConfigurationValue);
					}
				}
				
				/**
				 * Assign the value to the enumeration entry.
				 */
				configurationProperty.setValue(stringConfigurationValue);
				
				System.out.println(configurationProperty.name() + ": " + configurationProperty.getValue());
			}
			scanner.close();
		}
		
		/**
		 * Step 3: Create the time card data loader using the configuration properties
		 * enumeration.
		 */
		TimeCardDataLoader timeCardDataLoader = new TimeCardDataLoader(ConfigurationProperty.DataFilePath.getValue());
		
		/**
		 * Step 4: Load the time card data.
		 */
		timeCardDataLoader.loadFile();
	}
}