package com.bif.timecard.loader;

import java.text.SimpleDateFormat;

import com.bif.timecard.loader.enumeration.PayType;

/**
 * Instances of this class provide a object model for time card
 * records. As data files are loaded the data from each line is
 * converted into an instance of this class. When the loading is
 * done the collection of these objects is committed to the database.
 * @author <a href="mailto:roderick.barnes@biftechnologies.com">Roderick L. Barnes, Sr.</a>
 *
 */
public class TimeCardRecord {
	/**
	 * member attribute for the name of the file from which
	 * the time card data was loaded
	 */
	private String stringFileName;
	
	/**
	 * member attribute for the line number of the file from
	 * which the time card data was loaded
	 */
	private long longLineNumber;
	
	/**
	 * member attribute for the raw data that was converted into
	 * this record
	 */
	private String stringLineOfText;
	
	/**
	 * member attribute for the company identifier
	 */
	private String stringCompanyID;
	
	/**
	 * member attribute for the employee identifier
	 */
	private String stringEmployeeID;
	
	/**
	 * member attribute for the employee first name
	 */
	private String stringFirstName;
	
	/**
	 * member attribute for the employee last name
	 */
	private String stringLastName;
	
	/**
	 * member attribute for the date on which the event (meal, break,
	 * work, vacation, etc.) took place
	 */
	private String stringEventDate;
	
	/**
	 * member attribute for the pay type (event type) of the time card
	 * record
	 */
	private PayType payType;
	
	/**
	 * member attribute for the number of regular pay hours associated
	 * with the event
	 */
	private double doubleRegularDuration;

	/**
	 * member attribute for the number of unpaid hours associated
	 * with the event
	 */
	private double doubleUnpaidDuration;
	
	/**
	 * member attribute for the total number of hours for which the customer
	 * will be invoiced
	 */
	private double doubleTotalPaidDuration;
	
	/**
	 * member attribute for the project identifier
	 */
	private String stringProjectID;
	
	/**
	 * member attribute for the name of the project
	 */
	private String stringProjectName;
	
	/**
	 * member attribute for the labor category identifier
	 */
	private String stringLaborCategoryID;
	
	/**
	 * member attribute for the identifier of the department
	 */
	private String stringDepartmentID;
	
	/**
	 * member attribute for the name of the department in which the
	 * the employee was serving when the event took place
	 */
	private String stringDepartmentName;
	
	/**
	 * member attribute for the hourly rate of the employee
	 */
	private double doubleHourlyRate;
	
	/**
	 * Sets the name of the file from which the time card data was loaded.
	 * @param stringFileNameNew name of the file from which the time card
	 * data was loaded
	 */
	public void setFileName(String stringFileNameNew) {
		this.stringFileName = stringFileNameNew;
	}
	
	/**
	 * Returns the name of the file from which the time card data was loaded.
	 * @return the name of the file from which the time card data was loaded.
	 */
	public String getFileName() {
		return this.stringFileName;
	}
	
	/**
	 * Sets the line number from which the time card data was loaded
	 * @param longLineNumberNew line number from which the time card data was
	 * loaded
	 */
	public void setLineNumber(long longLineNumberNew) {
		this.longLineNumber = longLineNumberNew;
	}
	
	/**
	 * Returns the line number from which the time card data was loaded.
	 * @return the line number from which the time card data was loaded.
	 */
	public long getLineNumber() {
		return this.longLineNumber;
	}
	
	/**
	 * Sets the raw line of text from which the time card data was parsed
	 * @param stringLineOfTextNew the raw line of text from which the time card
	 * data was parsed.
	 */
	public void setLineOfText(String stringLineOfTextNew) {
		this.stringLineOfText = stringLineOfTextNew;
	}
	
	/**
	 * Returns the raw line of text from which the time card data was parsed.
	 * @return the raw line of text from which the time card data was parsed.
	 */
	public String getLineOfText() {
		return this.stringLineOfText;
	}
	
	/**
	 * Sets the company identifier attribute of the time card record.
	 * @param stringCompanyIDNew the company identifier attribute of the time
	 * card record.
	 */
	public void setCompanyID(String stringCompanyIDNew) {
		this.stringCompanyID = stringCompanyIDNew;
	}
	
	/**
	 * Returns the company identifier attribute of the time card record.
	 * @return the company identifier attribute of the time card record.
	 */
	public String getCompanyID() {
		return this.stringCompanyID;
	}
	
	/**
	 * Set the employee identifier for the time card record.
	 * @param stringEmployeeIDNew the employee identifier
	 */
	public void setEmployeeID(String stringEmployeeIDNew) {
		this.stringEmployeeID = stringEmployeeIDNew;
	}
	
	/**
	 * Returns the employee identifier for the time card record
	 * @return the employee identifier for the time card record
	 */
	public String getEmployeeID() {
		return this.stringEmployeeID;
	}
	
	/**
	 * Sets the last name of the employee for whom time and attendance
	 * data is being recorded. The identifier of the employee can be obtained
	 * through {@link #getEmployeeID()}.
	 * @param stringLastNameNew last name of the employee
	 */
	public void setLastName(String stringLastNameNew) {
		this.stringLastName = stringLastNameNew;
	}
	
	/**
	 * Returns the last name of the employee.
	 * @return the last name of the employee.
	 */
	public String getLastName() {
		return this.stringLastName;
	}
	
	/**
	 * Sets the first name of the employee for whom time and attendance
	 * data is being recorded. The identifier of the employee can be obtained
	 * through {@link #getEmployeeID()}.
	 * @param stringFirstNameNew first name of the employee
	 */
	public void setFirstName(String stringFirstNameNew) {
		this.stringFirstName = stringFirstNameNew;
	}
	
	/**
	 * Returns the first name of the employee.
	 * @return the first name of the employee.
	 */
	public String getFirstName() {
		return this.stringFirstName;
	}
	
	/**
	 * Sets the date on which the event (Meal, Work, Vacation, etc) took
	 * place
	 * @param stringEventDateNew the date on which the event took place
	 */
	public void setEventDate(String stringEventDateNew) {
		this.stringEventDate = stringEventDateNew;
	}
	
	/**
	 * Returns the date on which the event (Break, Meal, Work, Vacation,
	 * etc.) took place.
	 * @return the date on which the event took place.
	 */
	public String getEventDate() {
		return this.stringEventDate;
	}
	
	/**
	 * Sets the event type for the record.
	 * @param payTypeNew the event type for the record.
	 */
	public void setPayType(PayType payTypeNew) {
		this.payType = payTypeNew;
	}
	
	/**
	 * Returns the event type for the time card record.
	 * @return the event type for the time card record.
	 */
	public PayType getPayType() {
		return this.payType;
	}
	
	/**
	 * Sets the number of hours for which the customer will be charged.
	 * @param floatRegularDurationNew hours for which the customer will be charged.
	 */
	public void setRegularDuration(double floatRegularDurationNew) {
		this.doubleRegularDuration = floatRegularDurationNew;
	}
	
	/**
	 * Returns the number of hours for which the customer will be charged.
	 * @return the number of hours for which the customer will be charged.
	 */
	public double getRegularDuration() {
		return this.doubleRegularDuration;
	}
	
	/**
	 * Sets the number of hours in the event for which the customer will not
	 * be charged.
	 * @param floatUnpaidDurationNew the number of hours for which the customer
	 * will not be charged and for which the person will not be paid.
	 */
	public void setUnpaidDuration(double floatUnpaidDurationNew) {
		this.doubleUnpaidDuration = floatUnpaidDurationNew;
	}
	
	/**
	 * Returns the number of hours for which the custome will not be charged and
	 * for which the employee will not be paid.
	 * @return number of hours for which there will be no charge and no pay.
	 */
	public double getUnpaidDuration() {
		return this.doubleUnpaidDuration;
	}
	
	/**
	 * Sets the total number of hours for which the customer will be charged.
	 * @param floatRegularDurationNew total number of hours for which the customer 
	 * will be charged.
	 */
	public void setTotalPaidDuration(double doubleTotalPaidDurationNew) {
		this.doubleTotalPaidDuration = doubleTotalPaidDurationNew;
	}
	
	/**
	 * Returns the total number of hours for which the customer will be charged.
	 * @return the total number of hours for which the customer will be charged.
	 */
	public double getTotalPaidDuration() {
		return this.doubleTotalPaidDuration;
	}
	
	/**
	 * Sets the project identifier.
	 * @param stringProjectIDNew project identifier.
	 */
	public void setProjectID(String stringProjectIDNew) {
		this.stringProjectID = stringProjectIDNew;
	}
	
	/**
	 * Returns the project identifier.
	 * @return the project identifier.
	 */
	public String getProjectID() {
		return this.stringProjectID;
	}
	
	/**
	 * Sets the name of the project.
	 * @param stringProjectNameNew the name of the project.
	 */
	public void setProjectName(String stringProjectNameNew) {
		this.stringProjectName = stringProjectNameNew;
	}
	
	/**
	 * Returns the name of the project.
	 * @return the name of the project.
	 */
	public String getProjectName() {
		return this.stringProjectName;
	}
	
	/**
	 * Sets the labor category identifier for the work performed by the
	 * employee ({@link #getEmployeeID()}).
	 * @param stringLaborCategoryIDNew labor category identifier for the work.
	 */
	public void setLaborCategoryID(String stringLaborCategoryIDNew) {
		this.stringLaborCategoryID = stringLaborCategoryIDNew;
	}
	
	/**
	 * Returns the labor category identifier (e.g., 0001AL) for the work
	 * performed by the employee ({@link #getEmployeeID()}).
	 * @return
	 */
	public String getLaborCategoryID() {
		return this.stringLaborCategoryID;
	}
	
	/**
	 * Sets the department identifier in which the employee ({@link #getEmployeeID()}) performed
	 * the work ({@link #getLaborCategoryID()}).
	 * @param stringDepartmentIDNew
	 */
	public void setDepartmentID(String stringDepartmentIDNew) {
		this.stringDepartmentID = stringDepartmentIDNew;
	}
	
	/**
	 * Returns the department identifier in which the employee ({@link #getEmployeeID()}) performed
	 * the work ({@link #getLaborCategoryID()}).
	 * @return the department identifier in which the employee ({@link #getEmployeeID()}) performed
	 * the work ({@link #getLaborCategoryID()}).
	 */
	public String getDepartmentID() {
		return this.stringDepartmentID;
	}
	
	/**
	 * Name of the department in which the work ({@link #getLaborCategoryID()}) was performed.
	 * @param stringDepartmentNameNew name of the department.
	 */
	public void setDepartmentName(String stringDepartmentNameNew) {
		this.stringDepartmentName = stringDepartmentNameNew;
	}
	
	/**
	 * Returns the name of the department in which the work 
	 * ({@link #getLaborCategoryID()}) was performed.
	 * @return
	 */
	public String getDepartmentName() {
		return this.stringDepartmentName;
	}
	
	/**
	 * Sets the hourly rate for the work performed. This field is not persisted; the information
	 * can be derived from the labor category.
	 * @param doubleHourlyRateNew the hourly rate for the work performed.
	 */
	public void setHourlyRate(double doubleHourlyRateNew) {
		this.doubleHourlyRate = doubleHourlyRateNew;
	}
	
	/**
	 * Return the hourly rate for the work performed. This field is not persisted; the information
	 * can be derived from the labor category.
	 * @return the hourly rate for the work performed.
	 */
	public double getHourlyRate() {
		return this.doubleHourlyRate;
	}
	
	/**
	 * Returns the billing period that the time card record belongs in.
	 * @return the billing period that the time card record belongs in.
	 */
	public String getBillingPeriodID() {
		/**
		 * Step 1: Extract the year from the event date.
		 */
		String stringYear = this.getEventDate().split("\\/")[2];
		if (stringYear.length() == 2) {
			stringYear = "20" + stringYear;
		}
		
		/**
		 * Step 2: Extract the month from the event date.
		 */
		String stringMonth = this.getEventDate().split("\\/")[0];
		if (stringMonth.length() == 1) {
			stringMonth = "0" + stringMonth;
		}
		
		/**
		 * Step 3: Concatenate the year and month to get the billing period.
		 */
		return stringYear + "." + stringMonth;
	}
	
	/**
	 * Constructor for new instances of {@code TimeCardRecord}.
	 * @param stringFileNameNew name of the file from which the time card data was read
	 * @param longLineNumberNew line of the file from which the time card data was read
	 * @param stringLineOfTextNew line of text that was parsed to obtain the time card data
	 * @param stringCompanyIDNew company identifier (e.g., Google, Amazon, Microsoft)
	 * @param stringEmployeeIDNew employee identifier
	 * @param stringFirstNameNew first name of the employee
	 * @param stringLastNameNew last name of the employee
	 * @param stringEventDateNew date on which the event (Break, Meal, Work, Vaction) took place
	 * @param payTypeNew type of event (Break, Meal, Work, Vaction, Military Leave, etc.)
	 * @param doubleRegularDurationNew regular hours for which the customer will be invoiced
	 * @param doubleUnpaidDurationNew hours for which the custome will not be invoiced 
	 * 		(e.g. Break, Meal, Military Leave)
	 * @param doubleTotalPaidDurationNew total hours for which the customer will be invoiced
	 * @param stringProjectIDNew project identifier
	 * @param stringProjectNameNew name of the project
	 * @param stringLaborCategoryIDNew labor category identifier
	 * @param stringDepartmentIDNew department identifier
	 * @param StringDepartmentNameNew department name
	 * @param doubleHourlyRateNew hourly rate for the work (not persisted because it can be
	 * derived from the labor category).
	 */
	public TimeCardRecord(
			String stringFileNameNew,
			long longLineNumberNew,
			String stringLineOfTextNew,
			String stringCompanyIDNew,
			String stringEmployeeIDNew, 
			String stringFirstNameNew, 
			String stringLastNameNew, 
			String stringEventDateNew,
			PayType payTypeNew, 
			double doubleRegularDurationNew, 
			double doubleUnpaidDurationNew, 
			double doubleTotalPaidDurationNew, 
			String stringProjectIDNew, 
			String stringProjectNameNew, 
			String stringLaborCategoryIDNew, 
			String stringDepartmentIDNew,
			String StringDepartmentNameNew,
			double doubleHourlyRateNew
		) {
		this.setFileName(stringFileNameNew);
		this.setLineNumber(longLineNumberNew);
		this.setLineOfText(stringLineOfTextNew);
		this.setCompanyID(stringCompanyIDNew);
		this.setEmployeeID(stringEmployeeIDNew);
		this.setFirstName(stringFirstNameNew);
		this.setLastName(stringLastNameNew);
		this.setEventDate(stringEventDateNew);
		this.setPayType(payTypeNew);
		this.setRegularDuration(doubleRegularDurationNew);
		this.setUnpaidDuration(doubleUnpaidDurationNew);
		this.setTotalPaidDuration(doubleTotalPaidDurationNew);
		this.setProjectID(stringProjectIDNew);
		this.setProjectName(stringProjectNameNew);
		this.setLaborCategoryID(stringLaborCategoryIDNew);
		this.setDepartmentID(stringDepartmentIDNew);
		this.setDepartmentName(StringDepartmentNameNew);
		this.setHourlyRate(doubleHourlyRateNew);
	}
}